@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2004, 2005 Marcus Glocker <marcus@nazgul.ch>
+ * Copyright (c) 2004, 2005, 2006 Marcus Glocker <marcus@nazgul.ch>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -14,21 +14,26 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
+#include <signal.h>
+
 #include "config.h"
 
 /*
- * Typedefs
+ * typedefs
  */
-typedef unsigned short int	us_int;
+typedef unsigned short int us_int;
 
 /*
- * Memory mapped mimes types
+ * general variables
  */
-extern int mimes_size;
-extern char *mimes;
+extern volatile sig_atomic_t	quit;
+extern int			debug;
+extern int			mimes_size;
+extern char			*mimes;
+extern char			logn[1024];
 
 /*
- * Constant strings
+ * constant strings
  */
 extern const char *month[12];
 extern const char *day[7];
@@ -62,6 +67,7 @@ extern const char *http_fn_aut;
 extern const char *http_fn_auw;
 extern const char *http_fn_ran;
 extern const char *http_fn_rac;
+extern const char *http_fn_aen;
 extern const char *http_fv_srv;
 extern const char *http_fv_pro;
 extern const char *http_fv_cgi;
@@ -71,16 +77,17 @@ extern const char *http_fv_con_alv;
 extern const char *http_fv_con_cls;
 
 /*
- * Strings
+ * strings
  */
 extern char http_fv_alv[128];
 extern char http_fv_cte[128];
 extern char http_url[128];
+extern char http_urls[128];
 extern char http_sig[128];
 extern char http_path[128];
 
 /*
- * Server configuration values
+ * server configuration values
  */
 struct cfg {
 	int	serverport;		/* Listener Port */
@@ -96,8 +103,6 @@ struct cfg {
 	char	logaccess[1024];	/* Filename for access log */
 	char	docroot[1024];		/* Full Path to server document root */
 	char	docindex[1024];		/* Filename for index documents */
-	char	cgiroot[1024];		/* Full Path to cgi root */
-	char	cgialias[1024];		/* Alias for cgi calls */
 	char	user[1024];		/* Server runs under that user */
 	char	htaccess[1024];		/* Filename for basic auth */
 	char	htpasswd[1024];		/* Full Filename for basic auth users */
@@ -114,7 +119,7 @@ struct cfg {
 extern struct cfg config;
 
 /*
- * Information structure for header
+ * information structure for header
  */
 struct header {
 	char	rq_method[1024];	/* Request Method */
@@ -139,6 +144,7 @@ struct header {
 	char	rq_fv_hos[1024];	/* Request Value Host: */
 	char	rq_fv_aut[1024];	/* Request Value Authorization: */
 	char	rq_fv_ran[1024];	/* Request Value Range: */
+	char	rq_fv_aen[1024];	/* Request Value Accept-Encoding: */
 	char	rp_fv_dat[1024];	/* Response Value Date: */
 	char	rp_fv_dam[1024];	/* Response Value ? */
 	char	rp_fv_cte[1024];	/* Response Value Content-Type: */
@@ -149,6 +155,7 @@ struct header {
 	int	rp_foffs;		/* Response File Offset */
 	int	rp_hsize;		/* Response Header Size */
 	us_int	rp_status;		/* Response Status */
+	us_int	x_chk;			/* Flag for Chunking */
 	us_int	x_opt;			/* Flag for Option */
 	us_int	x_qry;			/* Flag for Query */
 	us_int	x_sct;			/* Flag for Script */
@@ -161,30 +168,38 @@ struct header {
 	us_int	x_hos;			/* Flag for Host: */
 	us_int	x_aut;			/* Flag for Authorization: */
 	us_int	x_ran;			/* Flag for Range: */
+	us_int	x_aen;			/* Flag for Range: */
 	us_int	x_cgi;			/* Flag for CGI */
 	us_int	x_idx;			/* Flag for Index */
 	us_int	x_dov;			/* Flag for Virutal Host */
 };
 
 /*
- * Information structure for each connection
+ * information structure for each connection
  */
 struct connection {
-	int	to;			/* Timeout */
-	us_int	state;			/* 0 = close, 1 = keep alive */
-	char	ip[24];			/* Client IP */
-	char	pt[8];			/* Client Port */
-	int	pfdo;			/* Partitial, Open File Desc */
-	int	pfdn[HDN];		/* Partitial, File Desc */
+	SSL	*ssl;			/* Connection, SSL Object */
+	char	pt[8];			/* Connection, Client Port */
+	char	ip[24];			/* Connection, Client IP */
+	char	*rbuf;			/* Connection, Read buf */
+	char	*wbuf;			/* Connection, Write buf */
+	int	roff;			/* Connection, Read buf offset */
+	int	wred;			/* Connection, Write buf read bytes */
+	int	wsnd;			/* Connection, Write buf sent bytes */
+	int	to;			/* Connection, Timeout */
+	int	pfdo;			/* Connection, Open File descriptors */
+	int	pfdn[HDN];		/* Partitial, File descriptor */
 	int	pfds[HDN];		/* Partitial, Bytes sent */
-	us_int	status[HDN];		/* Response status */
+	int	psta[HDN];		/* Partitial, Response Status */
+	char	*pfdh[HDN];		/* Partitial, Response Header */
 	char	*plreq[HDN];		/* Partitial, Log Request */
 	char	*plref[HDN];		/* Partitial, Log Referer */
 	char	*plagt[HDN];		/* Partitial, Log Client */
 	char	*pllog[HDN];		/* Partitial, Log File */
-	SSL	*ssl;			/* SSL Object */
-	us_int	x_ssl;			/* Flag for SSL Connection */
-	us_int	x_ip6;			/* Flag for IPv6 Connection */
-	us_int	x_chk[HDN];		/* Flag for Chunked Data */
+	us_int	x_chk[HDN];		/* Partitial, Flag for Chunked block */
+	us_int	x_ful[HDN];		/* Partitial, Flag for Full block */
+	us_int	x_ssl;			/* Connection, Flag for SSL */
+	us_int	x_ip6;			/* Connection, Flag for IPv6 */
+	us_int	x_sta;			/* Connection, Flag for State */
 };
 extern struct connection c[];
