@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2004, 2005 Marcus Glocker <marcus@nazgul.ch>
+ * Copyright (c) 2004, 2005, 2006 Marcus Glocker <marcus@nazgul.ch>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -48,16 +48,19 @@
 #include "extern.h"
 
 /*
- * Global vars local
+ * global vars local
  */
 static int		ppid;
 static const char	*configfile = "/var/nostromo/conf/nhttpd.conf";
 
 /*
- * Global vars extern
+ * global vars extern
  */
-int	mimes_size;
-char	*mimes;
+volatile sig_atomic_t	quit = 0;
+int			debug = 0;
+int			mimes_size;
+char			*mimes;
+char			logn[1024];
 
 const char *month[12] = {
 	"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct",
@@ -98,7 +101,8 @@ const char *http_fn_aut		= "Authorization:";
 const char *http_fn_auw		= "WWW-Authenticate:";
 const char *http_fn_ran		= "Range:";
 const char *http_fn_rac		= "Content-Range:";
-const char *http_fv_srv		= "Nostromo/1.7.5 (Unix)";
+const char *http_fn_aen		= "Accept-Encoding:";
+const char *http_fv_srv		= "nostromo 1.7.8";
 const char *http_fv_pro		= "HTTP/1.1";
 const char *http_fv_cgi		= "CGI/1.1";
 const char *http_fv_teg		= "chunked";
@@ -109,6 +113,7 @@ const char *http_fv_con_cls	= "close";
 char	http_fv_alv[128];
 char	http_fv_cte[128];
 char	http_url[128];
+char	http_urls[128];
 char	http_sig[128];
 char	http_path[128];
 
@@ -128,19 +133,22 @@ usage(int mode)
 
 	if (mode != 2) {
 		fprintf(stderr, "usage: %s ", __progname);
-		fprintf(stderr, "[-hrv46] [-c configfile]\n");
+		fprintf(stderr, "[-dhrv46] [-c configfile]\n");
 	}
+
 	if (mode == 1) {
 		fprintf(stderr, "\noptions:\n");
+		fprintf(stderr, "  -d\t\t: Debug mode.\n");
 		fprintf(stderr, "  -h\t\t: This help.\n");
-		fprintf(stderr, "  -v\t\t: Shows version.\n");
 		fprintf(stderr,
 		    "  -r\t\t: nhttpd will chroot to serverroot.\n");
+		fprintf(stderr, "  -v\t\t: Shows version.\n");
 		fprintf(stderr, "  -4\t\t: Enable IPv4 and IPv6.\n");
 		fprintf(stderr, "  -6\t\t: Enable IPv6 only.\n");
 		fprintf(stderr,
 		    "  -c configfile\t: Use an alternate configfile.\n");
 	}
+
 	if (mode == 2)
 		fprintf(stderr, "%s\n", http_fv_srv);
 
@@ -166,10 +174,7 @@ sig_handler(const int sig)
 	}
 	case SIGINT:
 	case SIGTERM:
-		/* terminate nhttpd gracefully */
-		munmap(mimes, mimes_size);
-		flogd(config.logserver, "[info] nhttpd stopped\n");
-		exit(0);
+		quit = 1;
 		break;
 	case SIGPIPE:
 		/* ignore */
@@ -190,6 +195,44 @@ sig_handler(const int sig)
 	}
 }
 
+/*
+ * send_full()
+ *	send a fully generated header/body block
+ * Return:
+ *	none
+ */
+void
+send_full(const int sfd, const int hr)
+{
+	int	s, len;
+	char	*offset;
+
+	s = 0;
+
+	len = strlen(c[sfd].pfdh[hr]) - c[sfd].wsnd;
+
+	if (len > 0) {
+		offset = c[sfd].pfdh[hr] + c[sfd].wsnd;
+
+		if (c[sfd].x_ssl)
+			s = sys_write_ssl(c[sfd].ssl, offset, len);
+		else
+			s = sys_write_a(sfd, offset, len);
+		if (s != -1)
+			c[sfd].wsnd += s;
+	}
+
+	if (len == 0 || s < 0) {
+		c[sfd].pfdo--;
+		c[sfd].wred = 0;
+ 		c[sfd].wsnd = 0;
+		c[sfd].pfdn[hr] = 0;
+		c[sfd].x_chk[hr] = 0;
+		c[sfd].x_ful[hr] = 0;
+		free(c[sfd].pfdh[hr]);
+	}
+}
+
 /*
  * send_chunk()
  *	send a file in chunks
@@ -199,40 +242,72 @@ sig_handler(const int sig)
 void
 send_chunk(const int sfd, const int hr)
 {
-	int		r, s;
-	char		*b;
-	static char	fbuf[BS];
+	int	k, h, s, len;
+	char	*offset;
 
-	r = s = 0;
+	k = h = s = 0;
 
-	/* we need to initialize buf to do nice chunks */
-	memset(fbuf, 0, sizeof(fbuf));
+	len = c[sfd].wred - c[sfd].wsnd;
 
 	/* read block from file */
-	r = read(c[sfd].pfdn[hr], fbuf, BS);
+	if (len == 0) {
+		if (c[sfd].pfds[hr] == 0)
+			h = strlen(c[sfd].pfdh[hr]);
 
-	/* send a chunk */
-	if (r > 0) {
-		b = http_chunk(fbuf);
-		if (c[sfd].x_ssl) {
-			s = SSL_write(c[sfd].ssl, b, strlen(b));
-		} else {
-			s = sys_write(sfd, b, strlen(b));
+		k = http_chunk_ovr(BS);
+
+		c[sfd].wred = sys_read(c[sfd].pfdn[hr], c[sfd].wbuf,
+		    (BS - h) - k);
+
+		/* chunk block */
+		if (c[sfd].wred > 0) {
+			offset = http_chunk(c[sfd].wbuf, c[sfd].wred);
+			c[sfd].wred += http_chunk_ovr(c[sfd].wred);
+
+			if (c[sfd].pfds[hr] == 0) {
+				/* attach http header */
+				h = strlcpy(c[sfd].wbuf, c[sfd].pfdh[hr], BS);
+				c[sfd].pfds[hr] = 1;
+			}
+
+			memcpy(c[sfd].wbuf + h, offset, c[sfd].wred);
+			free(offset);
+			c[sfd].wred += h;
 		}
-		free(b);
+
+		c[sfd].wsnd = 0;
+		len = c[sfd].wred;
 	}
 
-	/* send last chunk */
-	if (r < 1 || s < 1) {
-		if (c[sfd].x_ssl) {
-			s = SSL_write(c[sfd].ssl, http_fv_lch,
+	offset = c[sfd].wbuf + c[sfd].wsnd;
+
+	/* send a chunk */
+	if (c[sfd].wred > 0) {
+		if (c[sfd].x_ssl)
+			s = sys_write_ssl(c[sfd].ssl, offset, len);
+		else
+			s = sys_write_a(sfd, offset, len);
+		if (s != 1)
+			c[sfd].wsnd += s;
+	}
+
+	/* send last chunk, cleanup */
+	if (c[sfd].wred < 1 || s < 0) {
+		if (c[sfd].x_ssl)
+			s = sys_write_ssl(c[sfd].ssl, http_fv_lch,
 			    strlen(http_fv_lch));
-		} else {
-			s = sys_write(sfd, http_fv_lch, strlen(http_fv_lch));
-		}
+		else
+			s = sys_write_a(sfd, http_fv_lch,
+			    strlen(http_fv_lch));
+
 		close(c[sfd].pfdn[hr]);
-		c[sfd].x_chk[hr] = 0;
 		c[sfd].pfdo--;
+		c[sfd].wred = 0;
+	 	c[sfd].wsnd = 0;
+		c[sfd].pfdn[hr] = 0;
+		c[sfd].pfds[hr] = 0;
+		c[sfd].x_chk[hr] = 0;
+		free(c[sfd].pfdh[hr]);
 	}
 }
 
@@ -245,32 +320,53 @@ send_chunk(const int sfd, const int hr)
 void
 send_file(const int sfd, const int hr)
 {
-	int		r, s;
-	static char	fbuf[BS];
+	int	h, s, len;
+	char	*offset;
 
-	r = s = 0;
+	h = s = 0;
+
+	len = c[sfd].wred - c[sfd].wsnd;
 
 	/* read block from file */
-	r = read(c[sfd].pfdn[hr], fbuf, BS);
+	if (len == 0) {
+		if (c[sfd].pfds[hr] == 0) {
+			/* attach http header */
+			h = strlcpy(c[sfd].wbuf, c[sfd].pfdh[hr], BS);
+			c[sfd].wred = sys_read(c[sfd].pfdn[hr],
+			    c[sfd].wbuf + h, BS - h);
+			c[sfd].wred += h;
+		} else
+			c[sfd].wred = sys_read(c[sfd].pfdn[hr],
+			    c[sfd].wbuf, BS);
+
+		c[sfd].wsnd = 0;
+		len = c[sfd].wred;
+	}
+
+	offset = c[sfd].wbuf + c[sfd].wsnd;
 
 	/* send a block */
-	if (r > 0) {
-		if (c[sfd].x_ssl) {
-			s = SSL_write(c[sfd].ssl, fbuf, r);
-		} else {
-			s = sys_write(sfd, fbuf, r);
+	if (c[sfd].wred > 0) {
+		if (c[sfd].x_ssl)
+			s = sys_write_ssl(c[sfd].ssl, offset, len);
+		else
+			s = sys_write_a(sfd, offset, len);
+		if (s != -1) {
+			c[sfd].pfds[hr] += h ? s - h : s;
+			c[sfd].wsnd += s;
 		}
-		if (s != -1)
-			c[sfd].pfds[hr] += s;
 	}
 
 	/* file sent, write access log, cleanup */
-	if (r < 1 || s < 1) {
+	if (c[sfd].wred < 1 || s < 0) {
 		http_alog(sfd, hr);
 		close(c[sfd].pfdn[hr]);
+		c[sfd].pfdo--;
+		c[sfd].wred = 0;
+		c[sfd].wsnd = 0;
 		c[sfd].pfdn[hr] = 0;
 		c[sfd].pfds[hr] = 0;
-		c[sfd].pfdo--;
+		free(c[sfd].pfdh[hr]);
 		free(c[sfd].plreq[hr]);
 		free(c[sfd].plref[hr]);
 		free(c[sfd].plagt[hr]);
@@ -280,7 +376,7 @@ send_file(const int sfd, const int hr)
 
 /*
  * main()
- *	Nostromo webserver
+ *	nostromo webserver
  * Return:
  *	0 = terminated successfull, 1 = terminated with error
  */
@@ -288,12 +384,12 @@ int
 main(int argc, char *argv[])
 {
 	int			sdlis1, sdlis2, sdlis3, sdlis4;
-	int			sdmax, sdnew, sdnow;
-	int			i, j, k, r, s, ch, rd, rp, rt, wr, post, size;
+	int			sdmax, sdnew, sdnow, sdlim;
+	int			i, j, k, r, s, ch, rd, rt, wr, post, size;
 	int			fd, erro, opt_c, opt_r, opt_4, opt_6;
 	char			in[HDN * HDS + 1], tmp[HDS + 1];
 	char			header[HDN][HDS + 1];
-	char			*b, *h, *body = NULL;
+	char			*b, *h, *offset, *body = NULL;
 	extern char		**environ;
 	struct passwd		*pwd = NULL;
 	struct stat		st;
@@ -303,21 +399,24 @@ main(int argc, char *argv[])
 	struct in6_addr		in6addr_any = IN6ADDR_ANY_INIT;
 	struct timeval		tv;
 	socklen_t		ssize;
-	fd_set			master_r, master_w, set_r, set_w, seto_r;
+	fd_set			master_r, master_w, set_r, set_w;
 	time_t			tnow;
 	SSL_CTX			*ssl_ctx = NULL;
 	FILE			*file;
 
-	r = s = rd = sdmax = 0;
+	i = r = s = rd = sdmax = 0;
 	sdlis1 = sdlis2 = sdlis3 = sdlis4 = -1;
 
 	/*
-	 * Command line arguments
+	 * command line arguments
 	 */
 	opt_4 = 1;
 	opt_6 = opt_c = opt_r = 0;
-	while ((ch = getopt(argc, argv, "hrv46c:")) != -1) {
+	while ((ch = getopt(argc, argv, "dhrv46c:")) != -1) {
 		switch (ch) {
+		case 'd':
+			debug = 1;
+			break;
 		case 'h':
 			usage(1);
 			break;
@@ -353,8 +452,9 @@ main(int argc, char *argv[])
 		close(fd);
 
 	/*
-	 * Parse configuration file
+	 * parse configuration file
 	 */
+
 	/* mandatory configuration parameters */
 	if (fparse(config.servername, "servername", config.file,
 	    sizeof(config.servername)) == -1) {
@@ -368,10 +468,10 @@ main(int argc, char *argv[])
 			config.serverport = PRT;
 			snprintf(config.serverportc, sizeof(config.serverportc),
 			    "%d", PRT);
-		} else {
+		} else
 			/* custom port */
 			config.serverport = atoi(config.serverportc);
-		}
+
 		/* remove port from servername */
 		strcuts(config.servername, config.servername, '\0', ':',
 		    sizeof(config.servername));
@@ -405,7 +505,8 @@ main(int argc, char *argv[])
 	    sizeof(config.logaccess)) == -1) {
 		fprintf(stderr, "Missing <logaccess> in config file.\n");
 		r = -1;
-	}
+	} else
+		strlcpy(logn, config.logaccess, sizeof(logn));
 	if (fparse(config.docroot, "docroot", config.file,
 	    sizeof(config.docroot)) == -1) {
 		fprintf(stderr, "Missing <docroot> in config file.\n");
@@ -416,16 +517,7 @@ main(int argc, char *argv[])
 		fprintf(stderr, "Missing <docindex> in config file.\n");
 		r = -1;
 	}
-	if (fparse(config.cgiroot, "cgiroot", config.file,
-	    sizeof(config.cgiroot)) == -1) {
-		fprintf(stderr, "Missing <cgiroot> in config file.\n");
-		r = -1;
-	}
-	if (fparse(config.cgialias, "cgialias", config.file,
-	    sizeof(config.cgialias)) == -1) {
-		fprintf(stderr, "Missing <cgialias> in config file.\n");
-		r = -1;
-	}
+
 	/* optional configuration parameters */
 	if (fparse(config.user, "user", config.file, sizeof(config.user))
 	    == -1)
@@ -437,9 +529,9 @@ main(int argc, char *argv[])
 	    sizeof(config.htpasswd)) == -1)
 		strlcpy(config.htpasswd, "0", sizeof(config.htpasswd));
 	if (fparse(config.sslportc, "sslport", config.file,
-	    sizeof(config.sslportc)) == -1) {
+	    sizeof(config.sslportc)) == -1)
 		config.ssl = 0;
-	} else {
+	else {
 		config.ssl = 1;
 		config.sslport = atoi(config.sslportc);
 		if (fparse(config.sslcert, "sslcert", config.file,
@@ -463,18 +555,19 @@ main(int argc, char *argv[])
 	if (fparse(config.c404, "custom_404", config.file, sizeof(config.user))
 	    == -1)
 		strlcpy(config.c404, "0", sizeof(config.c404));
+
 	/* we need all mandatory parameters otherwise no start */
 	if (r == -1)
 		exit(1);
 
 	/*
-	 * Change to serverroot directory
+	 * change to serverroot directory
 	 */
 	if (chdir(config.serverroot) == -1)
 		err(1, "chdir");
 
 	/*
-	 * Get UID and GID
+	 * get UID and GID
 	 */
 	if (config.user[0] != '0') {
 		if ((pwd = getpwnam(config.user)) == NULL)
@@ -482,13 +575,13 @@ main(int argc, char *argv[])
 	}
 
 	/*
-	 * Map mime types file
+	 * map mime types file
 	 */
 	stat(config.servermimes, &st);
 	mimes_size = st.st_size;
-	if ((fd = open(config.servermimes, O_RDONLY, 0)) == -1) {
+	if ((fd = open(config.servermimes, O_RDONLY, 0)) == -1)
 		err(1, config.servermimes);
-	} else {
+	else {
 		if ((mimes = mmap(NULL, mimes_size, PROT_READ,
 		    MAP_FILE|MAP_SHARED, fd, 0)) == MAP_FAILED)
 			err(1, "mmap");
@@ -496,15 +589,20 @@ main(int argc, char *argv[])
 	}
 
 	/*
-	 * Assemble some data
+	 * assemble some strings
 	 */
+
 	/* servers full url */
-	if (config.serverport == PRT) {
+	if (config.serverport == PRT)
 		snprintf(http_url, sizeof(http_url), "%s", config.servername);
-	} else {
-		snprintf(http_url, sizeof(http_url), "%s:%d", config.servername,
-		    config.serverport);
-	}
+	else
+		snprintf(http_url, sizeof(http_url), "%s:%d",
+		    config.servername, config.serverport);
+	if (config.ssl && config.sslport == PRTS)
+		snprintf(http_urls, sizeof(http_url), "%s", config.servername);
+	if (config.ssl && config.sslport != PRTS)
+		snprintf(http_urls, sizeof(http_url), "%s:%d",
+		    config.servername, config.sslport);
 	/* servers connection timeout */
 	snprintf(http_fv_alv, sizeof(http_fv_alv), "timeout=%d, max=%d", TON,
 	    TON);
@@ -512,13 +610,13 @@ main(int argc, char *argv[])
 	sys_mime(http_fv_cte, sizeof(http_fv_cte), mimes, mimes_size, "html");
 
 	/*
-	 * Clear all environment variables and save PATH
+	 * clear all environment variables and save PATH
 	 */
 	strlcpy(http_path, getenv("PATH"), sizeof(http_path));
 	memset(environ, 0, sizeof(environ));
 
 	/*
-	 * Get servers IP adress(es)
+	 * get servers IP adress(es)
 	 */
 	if (opt_4) {
 		if ((host = gethostbyname2(config.servername, AF_INET)) ==
@@ -563,7 +661,7 @@ main(int argc, char *argv[])
 	}
 
 	/*
-	 * Initialize select sets
+	 * initialize select sets
 	 */
 	FD_ZERO(&master_r);
 	FD_ZERO(&master_w);
@@ -729,7 +827,12 @@ main(int argc, char *argv[])
 	}
 
 	/*
-	 * Signal Handler
+	 * calculate maximal concurrent connections
+	 */
+	sdlim = CON + sdmax;
+
+	/*
+	 * signal handler
 	 */
 	signal(SIGCHLD, sig_handler);
 	signal(SIGTERM, sig_handler);
@@ -740,9 +843,8 @@ main(int argc, char *argv[])
 	signal(SIGALRM, sig_handler);
 
 	/*
-	 * Daemonize
+	 * daemonize
 	 */
-	ppid = getpid();
 	ppid = fork();
 	if (ppid == -1)
 		err(1, "fork");
@@ -767,7 +869,7 @@ main(int argc, char *argv[])
 	dup(0);
 
 	/*
-	 * Chroot to serverroot directory
+	 * chroot to serverroot directory
 	 */
 	if (opt_r) {
 		if (chroot(config.serverroot) == -1) {
@@ -780,7 +882,7 @@ main(int argc, char *argv[])
 	}
 
 	/*
-	 * Set UID and GID
+	 * set UID and GID
 	 */
 	if (config.user[0] != '0') {
 		if (setgid(pwd->pw_gid) == -1) {
@@ -813,7 +915,7 @@ main(int argc, char *argv[])
 	}
 
 	/*
-	 * Nostromo ready
+	 * nostromo ready
 	 */
 	flogd(config.logserver, "[info] nhttpd started\n");
 	if (config.ssl)
@@ -826,11 +928,13 @@ main(int argc, char *argv[])
 		flogd(config.logserver, "[info] IPv4 and IPv6 enabled\n");
 	if (!opt_4 && opt_6)
 		flogd(config.logserver, "[info] IPv6 only enabled\n");
+	if (debug)
+		flogd(config.logserver, "[info] debug mode enabled\n");
 
 	/*
-	 * Main loop
+	 * main loop
 	 */
-	while (1) {
+	while (!quit) {
 		/* copy master set */
 		set_r = master_r;
 		set_w = master_w;
@@ -839,7 +943,9 @@ main(int argc, char *argv[])
 		tv.tv_sec = TON;
 		tv.tv_usec = 0;
 		if (select(sdmax + 1, &set_r, &set_w, NULL, &tv) == -1) {
-			if (errno != EINTR)
+			if (errno == EINTR)
+				usleep(1);
+			else
 				flogd(config.logserver, "[error] select: %s\n",
 				    strerror(errno));
 			continue;
@@ -860,20 +966,22 @@ main(int argc, char *argv[])
 			wr = FD_ISSET(sdnow, &set_w);
 
 			/*
-			 * Existing connection idles, check for timeout
+			 * existing connection idles, check for timeout
 			 */
 			if (rd == 0 && wr == 0) {
 				if (sdnow != sdlis1 && sdnow != sdlis2 &&
 				    sdnow != sdlis3 && sdnow != sdlis4 &&
 				    c[sdnow].pfdo == 0 &&
 				    (time(&tnow) - c[sdnow].to) > TON) {
+					sys_log(config.logserver, debug,
+					"main: socket closed by us: timeout\n");
 					goto quit;
 				}
 				continue;
 			}
 
 			/*
-			 * New connection
+			 * new connection
 			 */
 			if (rd > 0 && (sdnow == sdlis1 || sdnow == sdlis2 ||
 			    sdnow == sdlis3 || sdnow == sdlis4)) {
@@ -893,13 +1001,28 @@ main(int argc, char *argv[])
 				}
 
 				/* set socket options */
-				if (fcntl(sdnew, F_SETFL, fcntl(sdnew,
-				    F_GETFL) | O_NONBLOCK) == -1)
-					continue;
-				size = HDN * BS;
-				if (setsockopt(sdnew, SOL_SOCKET, SO_SNDBUF,
-				    &size, sizeof(size)) == -1)
+				if (fcntl(sdnew, F_SETFL,
+				    fcntl(sdnew, F_GETFL) | O_NONBLOCK) == -1)
 					continue;
+				if (SBS > 0) {
+					size = SBS;
+					if (setsockopt(sdnew, SOL_SOCKET,
+					    SO_SNDBUF, &size, sizeof(size)) ==
+					    -1)
+						continue;
+				}
+
+				/* initialize connection structure */
+				c[sdnew].to = time(&tnow);
+				c[sdnew].rbuf = malloc((HDN * HDS) + 1);
+				c[sdnew].wbuf = malloc(BS);
+				c[sdnew].roff = 0;
+				c[sdnew].wred = 0;
+				c[sdnew].wsnd = 0;
+				c[sdnew].pfdo = 0;
+				c[sdnew].x_ssl = 0;
+				c[sdnew].x_ip6 = 0;
+				c[sdnew].x_sta = 0;
 
 				/* SSL ready */
 				if (sdnow == sdlis2 || sdnow == sdlis4) {
@@ -926,22 +1049,19 @@ main(int argc, char *argv[])
 				}
 
 				/* check connection limit */
-				if (sdnew > CON) {
+				if (sdnew > sdlim) {
 					h = http_head(http_s_503, "-",
-					    c[sdnew].ip);
-					b = http_body(http_s_503, "", h);
-					if (c[sdnew].x_ssl) {
-						SSL_write(c[sdnew].ssl, b,
-						    strlen(b));
-						SSL_free(c[sdnew].ssl);
-						c[sdnew].x_ssl = 0;
-					} else {
-						sys_write(sdnew, b, strlen(b));
-					}
+					    c[sdnew].ip, 0);
+					b = http_body(http_s_503, "", h, 0);
+					c[sdnew].pfdo++;
+					c[sdnew].pfdn[i] = 1;
+					c[sdnew].pfdh[i] = strdup(b);
+					c[sdnew].x_ful[i] = 1;
+					c[sdnew].x_chk[i] = 0;
+					c[sdnew].x_sta = 0;
 					free(h);
 					free(b);
-					close(sdnew);
-					continue;
+					FD_SET(sdnew, &master_w);
 				}
 
 				/* add new connection to read fdset */
@@ -951,9 +1071,6 @@ main(int argc, char *argv[])
 				if (sdnew > sdmax)
 					sdmax = sdnew;
 
-				/* set timestamp for connection timeout */
-				c[sdnew].to = time(&tnow);
-
 				continue;
 			}
 
@@ -976,103 +1093,101 @@ main(int argc, char *argv[])
 			}
 
 			/*
-			 * Active connection wants partial file send
+			 * active connection wants partial file send
 			 */
 			if (c[sdnow].pfdo > 0) {
 				for (i = 0; i < HDN; i++) {
 					if (c[sdnow].pfdn[i] == 0)
 						continue;
-					/* send data chunked */
-					if (c[sdnow].x_chk[i])
-						send_chunk(sdnow, i);
-					/* send data block */
-					if (!c[sdnow].x_chk[i])
+
+					/* send plain block */
+				 	if (c[sdnow].x_chk[i] == 0 &&
+					    c[sdnow].x_ful[i] == 0)
 						send_file(sdnow, i);
+					/* send chunked block */
+					if (c[sdnow].x_chk[i] == 1 &&
+					    c[sdnow].x_ful[i] == 0)
+						send_chunk(sdnow, i);
+					/* send full response block */
+					if (c[sdnow].x_ful[i] == 1 &&
+					    c[sdnow].x_chk[i] == 0)
+						send_full(sdnow, i);
+
+					break;
 				}
 				/* all partial files sent */
 				if (c[sdnow].pfdo == 0) {
 					FD_CLR(sdnow, &master_w);
-					if (c[sdnow].state == 0)
+					if (c[sdnow].x_sta == 0)
 						goto quit;
 				}
 				continue;
 			}
 
 			/*
-			 * Active connection sends request
+			 * active connection sends request
 			 */
 			memset(tmp, 0, sizeof(tmp));
 
+			/* size check for data */
+ 			if (c[sdnow].roff > (sizeof(in) - 1) - BS) {
+				h = http_head(http_s_413, "-", c[sdnow].ip, 0);
+				b = http_body(http_s_413, "", h, 0);
+				c[sdnow].pfdo++;
+				c[sdnow].pfdn[i] = 1;
+				c[sdnow].pfdh[i] = strdup(b);
+				c[sdnow].x_ful[i] = 1;
+				c[sdnow].x_chk[i] = 0;
+       				c[sdnow].x_sta = 0;
+				free(h);
+				free(b);
+				continue;
+			}
+
 			/* receive data */
+			offset = c[sdnow].rbuf + c[sdnow].roff;
+
 			if (c[sdnow].x_ssl)
-				rt = SSL_read(c[sdnow].ssl, in, sizeof(in) - 1);
+				r = sys_read_ssl(c[sdnow].ssl, offset, BS);
 			else
-				rt = sys_read(sdnow, in, sizeof(in) - 1);
-			if (rt < 1)
+				r = sys_read_a(sdnow, offset, BS);
+			if (r < 1)
 				goto quit;
-			in[rt] = '\0';
+
+			c[sdnow].roff += r;
+			offset = c[sdnow].rbuf + c[sdnow].roff;
+			*offset = '\0';
 
 			/* set timestamp for connection timeout */
 			c[sdnow].to = time(&tnow);
 
-			/* receive everything */
-			FD_ZERO(&seto_r);
-			FD_SET(sdnow, &seto_r);
-			while ((size = http_header_comp(in)) == 0) {
-				tv.tv_sec = 1;
-				tv.tv_usec = 0;
-				r = select(sdnow + 1, &seto_r, NULL, NULL, &tv);
-				if (r == -1) {
-					if (errno == EINTR)
-						continue;
-					goto quit;
-				}
-				if (r == 0)
-					goto quit;
-				/* size check for data */
-				if (rt == sizeof(in) - 1) {
-					h = http_head(http_s_413, "-",
-					    c[sdnow].ip);
-					b = http_body(http_s_413, "", h);
-					if (c[sdnow].x_ssl) {
-						SSL_write(c[sdnow].ssl, b,
-						    strlen(b));
-					} else {
-						sys_write(sdnow, b, strlen(b));
-					}
-					free(h);
-					free(b);
-					goto quit;
-				}
-				if (c[sdnow].x_ssl) {
-					rp = SSL_read(c[sdnow].ssl, in + rt,
-					    (sizeof(in) - rt) - 1);
-				} else {
-					rp = sys_read(sdnow, in + rt,
-					    (sizeof(in) - rt) - 1);
-				}
-				if (rp < 1)
-					goto quit;
-				rt += rp;
-			}
-			in[rt] = '\0';
+			/* we need a full header sequence */
+			size = http_header_comp(c[sdnow].rbuf);
+			if (!size)
+				continue;
+
+			/* we got a full header sequence */
+			strlcpy(in, c[sdnow].rbuf, sizeof(in));
+			rt = c[sdnow].roff;
+			c[sdnow].roff = 0;
 
 			/* size check for header count */
 			if (size > HDN) {
-				h = http_head(http_s_413, "-", c[sdnow].ip);
-				b = http_body(http_s_413, "", h);
-				if (c[sdnow].x_ssl) {
-					SSL_write(c[sdnow].ssl, b, strlen(b));
-				} else {
-					sys_write(sdnow, b, strlen(b));
-				}
+				h = http_head(http_s_413, "-", c[sdnow].ip, 0);
+				b = http_body(http_s_413, "", h, 0);
+				c[sdnow].pfdo++;
+				c[sdnow].pfdn[i] = 1;
+				c[sdnow].pfdh[i] = strdup(b);
+				c[sdnow].x_ful[i] = 1;
+				c[sdnow].x_chk[i] = 0;
+       				c[sdnow].x_sta = 0;
 				free(h);
 				free(b);
-				goto quit;
+				continue;
 			}
 
 			/* post */
-			if (strncmp("POST ", in, 5) == 0)
+			if (!strncasecmp("POST ", in, 5))
 				post = 1;
 			else
 				post = 0;
@@ -1082,17 +1197,17 @@ main(int argc, char *argv[])
 				/* size check for header */
 				if (j == sizeof(header[k]) - 1) {
 					h = http_head(http_s_413, "-",
-					    c[sdnow].ip);
-					b = http_body(http_s_413, "", h);
-					if (c[sdnow].x_ssl) {
-						SSL_write(c[sdnow].ssl, b,
-						    strlen(b));
-					} else {
-						sys_write(sdnow, b, strlen(b));
-					}
+					    c[sdnow].ip, 0);
+					b = http_body(http_s_413, "", h, 0);
+					c[sdnow].pfdo++;
+					c[sdnow].pfdn[i] = 1;
+					c[sdnow].pfdh[i] = strdup(b);
+					c[sdnow].x_ful[i] = 1;
+					c[sdnow].x_chk[i] = 0;
+       					c[sdnow].x_sta = 0;
 					free(h);
 					free(b);
-					goto quit;
+					continue;
 				}
 				if (in[i] == '\n' && in[i + 2] == '\n') {
 					tmp[j] = in[i];
@@ -1119,25 +1234,29 @@ main(int argc, char *argv[])
 
 			/* and process every single header */
 			for (i = 0; i < k; i++) {
-				r = http_verify(header[i], c[sdnow].ip, sdnow);
-				if (r == 1)
+				s = 0;
+				r = 1;
+				s = http_verify(header[i], c[sdnow].ip, sdnow,
+				    i);
+				if (s == 1)
 					r = http_proc(header[i], body, i, size,
 					    sdnow);
+				if (r == 0)
+					goto quit;
 				/* on error abort processing */
 				if (r == -1) {
 					strcutl(tmp, header[i], 1, sizeof(tmp));
 					h = http_head(http_s_500, tmp,
-					    c[sdnow].ip);
-					b = http_body(http_s_500, "", h);
-					if (c[sdnow].x_ssl) {
-						SSL_write(c[sdnow].ssl, b,
-						    strlen(b));
-					} else {
-						sys_write(sdnow, b, strlen(b));
-					}
+					    c[sdnow].ip, 0);
+					b = http_body(http_s_500, "", h, 0);
+					c[sdnow].pfdo++;
+					c[sdnow].pfdn[i] = 1;
+					c[sdnow].pfdh[i] = strdup(b);
+					c[sdnow].x_ful[i] = 1;
+					c[sdnow].x_chk[i] = 0;
+       					c[sdnow].x_sta = 0;
 					free(h);
 					free(b);
-					goto quit;
 				}
 			}
 
@@ -1153,11 +1272,20 @@ main(int argc, char *argv[])
 				c[sdnow].to = 0;
 				c[sdnow].x_ip6 = 0;
 				c[sdnow].x_ssl = 0;
+				free(c[sdnow].rbuf);
+				free(c[sdnow].wbuf);
 				close(sdnow);
 				FD_CLR(sdnow, &master_r);
 			}
 		}
 	}
 
-	return 0;
+	/*
+	 * nostromo shutdown
+	 */
+	munmap(mimes, mimes_size);
+	flogd(config.logserver, "[info] nhttpd stopped\n");
+	kill(0, SIGTERM);	/* terminate all childs */
+
+	return (0);
 }
