@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2004, 2005 Marcus Glocker <marcus@nazgul.ch>
+ * Copyright (c) 2004, 2005, 2006 Marcus Glocker <marcus@nazgul.ch>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -19,13 +19,14 @@
  */
 void		usage(const int);
 void		sig_handler(const int);
+void		send_full(const int, const int);
 void		send_chunk(const int, const int);
-void		send_block(const int, const int);
+void		send_file(const int, const int);
 
 /*
  * http.c
  */
-int		http_verify(const char *, const char *, const int);
+int		http_verify(const char *, const char *, const int, const int);
 int		http_proc(const char *, char *, const int, const int,
 		    const int);
 int		http_cgi_getexec(char *, char *, const char *, const int,
@@ -35,11 +36,12 @@ int		http_header_comp(char *);
 int		http_body_comp(char *, const int, const int, const int);
 int		http_access_htaccess(char *, const char *, const int);
 int		http_alog(const int, const int);
-char *		http_chunk(const char *);
+int		http_chunk_ovr(const int);
+char *		http_chunk(const char *, const int);
 char *		http_date(struct tm *);
 char *		http_uridecode(const char *);
-char *		http_head(const char *, const char *, const char *);
-char *		http_body(const char *, const char *, const char *);
+char *		http_head(const char *, const char *, const char *, const int);
+char *		http_body(const char *, const char *, const char *, const int);
 struct header *	http_header(const char *, const char *, const int, const int,
 		    const int);
 
@@ -50,7 +52,13 @@ int		sys_mime(char *, const int, const char *, const int,
 		    const char *);
 int		sys_access_auth(const char *, const char *);
 int		sys_read(const int, char *, const int);
+int		sys_read_a(const int, char *, const int);
+int		sys_read_ssl(SSL *, char *, const int);
 int		sys_write(const int, const char *, const int);
+int		sys_write_a(const int, const char *, const int);
+int		sys_write_ssl(SSL *, const char *, const int);
+int		sys_resetnonblock(const int);
 int		sys_close_except(const int);
+int		sys_log(const char *, const int, const char *, ...);
 char *		sys_date(struct tm *);
 char *		sys_benv(const char *, ...);
